# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from collections import deque
from copy import deepcopy
from typing import Any, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union

import nncf
import nncf.tensor.functions as fns
from nncf.common.tensor import TensorType
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.experimental.common.tensor_statistics.statistical_functions import mean_per_channel
from nncf.experimental.common.tensor_statistics.statistics import MedianMADTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import TensorStatistic
from nncf.quantization.advanced_parameters import AggregatorType
from nncf.quantization.range_estimator import StatisticsType
from nncf.tensor import Tensor

InplaceInsertionFNType = TypeVar("InplaceInsertionFNType")
AggregationAxes = Tuple[int, ...]


class TensorReducerBase(ABC):
    """
    Tensor reducer is a callable object that reduces tensors according to
    the specified rule. Could handle tensors inplace or out of place.
    """

    def __init__(self, reduction_axes: Optional[ReductionAxes] = None, inplace: bool = False):
        """
        :param reduction_axes: Reduction axes for reduction calculation. Equal to list(range(len(input.shape)))
            if empty.
        :param inplace: Whether should be calculated inplace or out of place.
        """
        self._reduction_axes = reduction_axes
        self._inplace = inplace
        self._keepdims = True

    @property
    def inplace(self):
        return self._inplace

    @property
    def output_port_id(self) -> int:
        """
        Port id of the last node of the reducer subgraph if statistic is inplace.
        Port id of the reducer output return node if statistic is not inplace.
        """
        return 0

    @property
    def name(self):
        return self.__class__.__name__ + str(self.__hash__())

    @abstractmethod
    def _reduce_out_of_place(self, x: List[TensorType]) -> List[TensorType]:
        """
        Specifies the reduction rule.

        :param x: Tensor to register.
        """

    def get_inplace_fn(self) -> Optional[InplaceInsertionFNType]:
        """
        Returns correspondent inplace operation builder if inplace operations are available in backend.

        :return: Inplace operation builder if possible else None.
        """
        return None

    def __call__(self, x: List[Tensor]):
        if any(t.isempty() for t in x):
            return None

        if self.inplace:
            return x

        return self._reduce_out_of_place(x)

    def __eq__(self, __o: object) -> bool:
        return (
            isinstance(__o, self.__class__)
            and self._reduction_axes == __o._reduction_axes
            and self._inplace == __o.inplace
        )

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.inplace, self._reduction_axes))

    def _get_reduction_axes(self, tensor: Tensor) -> ReductionAxes:
        if self._reduction_axes is not None:
            return self._reduction_axes
        return tuple(range(len(tensor.shape)))


class AggregatorBase:
    """
    Aggregator is designed to receive (register) calculated statistics and aggregate them.
    """

    def __init__(
        self,
        aggregation_axes: Optional[AggregationAxes] = None,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ):
        """
        :param aggregation_axes: Axes along which to operate.
            Registered statistics are stacked along zero axis,
            axes >=1 correspond to received statistic axes shifted left by 1.
        :param num_samples: Maximum number of samples to collect. Aggregator
            skips tensor registration if tensor registration was called num_samples times before.
            Aggregator never skips registration if num_samples is None.
        :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
        """

        self._aggregation_axes = (0,) if aggregation_axes is None else aggregation_axes
        self._keepdims = True
        self._num_samples = num_samples
        self._collected_samples = 0
        self._window_size = window_size
        self._container = deque(maxlen=window_size)

    @property
    def num_samples(self) -> int:
        return self._num_samples

    def register_reduced_input(self, x: TensorType):
        if self._num_samples is not None and self._collected_samples >= self._num_samples:
            return
        self._register_reduced_input_impl(x)
        self._collected_samples += 1

    @abstractmethod
    def _register_reduced_input_impl(self, x: TensorType) -> None:
        """
        Registers incoming tensor in tensor aggregator.

        :param x: Tensor to register.
        """

    def aggregate(self) -> Any:
        """
        Aggregates collected tensors and returns aggregated result.
        In case no tensors were collected returns None.

        :return: Aggregated result.
        """
        if self._collected_samples:
            return self._aggregate_impl()
        return None

    @abstractmethod
    def _aggregate_impl(self) -> Any:
        """
        Aggregates collected tensors and returns aggregated result.

        :return: Aggregated result.
        """

    def reset(self):
        self._collected_samples = 0
        self._container = []

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, self.__class__) and self._num_samples == __o.num_samples

    def __hash__(self) -> int:
        return hash(self.__class__.__name__)


class TensorCollector:
    """
    Calculates statistics at given tensors according to registered statistic branches.
    Statistic branch consists of one reducer and one aggregator instance. TensorCollector
    applies a reducer on a correspondent inputs and then passes the one of the reduced tensors
    chosen by output port id to a correspondent aggregator for each registered statistic branch.
    Receives tensors by `register_input` method. Aggregated values as a TensorStatistic instance or
    a dict could be collected by `get_statistics` call.
    """

    def __init__(self, statistic_container: Optional[Type[TensorStatistic]] = None) -> None:
        self._reducers: Set[TensorReducerBase] = set()
        self._aggregators: Dict[Tuple[int, int, int], AggregatorBase] = {}
        self._stat_container_kwargs_map: Dict[str, Tuple[int, int, int]] = {}
        self._stat_container = statistic_container
        self.enable()
        self.clear_cache()

    @property
    def num_samples(self) -> Optional[int]:
        output = None
        for aggregator in self._aggregators.values():
            if aggregator.num_samples and output:
                output = max(output, aggregator.num_samples)
            else:
                output = aggregator.num_samples
        return output

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def reducers(self):
        return self._reducers.copy()

    @property
    def aggregators(self):
        return self._aggregators.copy()

    def enable(self):
        self._enabled = True

    def disable(self):
        self._enabled = False

    def register_statistic_branch(
        self,
        container_key: str,
        reducer: TensorReducerBase,
        aggregator: AggregatorBase,
        reducer_output_port_id: int = 0,
    ) -> None:
        """
        Registers statistic collection branch for a container key. Correspondent input will be reduced
        by given reducer and reduced value will be registered and aggregated by given aggregator.
        Passed container key should be unique for the TensorCollector instance.
        Passed aggregator instance should never be used twice for one TensorCollector instance.

        :param container_key: Container key to pass aggregated statistic to.
        :param reducer: TensorReducer instance for the statistic collection branch.
        :param aggregator: TensorAggregator instance for the statistic collection branch.
        :reducer_output_port_id: Reducer target output port id.
        """
        if container_key in self._stat_container_kwargs_map:
            raise nncf.InternalError(
                f"Two different statistic branches for one container key {container_key} are encountered"
            )
        if any(aggr is aggregator for aggr in self._aggregators.values()):
            raise nncf.InternalError(f"One aggregator instance {aggregator} for different branches is encountered")

        self._reducers.add(reducer)
        key = (hash(reducer), reducer_output_port_id, hash(aggregator))

        if key not in self._aggregators:
            self._aggregators[key] = aggregator
        self._stat_container_kwargs_map[container_key] = key

    def register_inputs(self, inputs: Dict[int, List[Tensor]]) -> None:
        """
        Registers given input in TensorCollector.

        :param inputs: Tensor inputs in format of dict where keys
            are reducer names and values are correspondent input tensors
        """
        if not self.enabled:
            return
        reduced_inputs = {}
        for reducer in self._reducers:
            reducer_hash = hash(reducer)
            input_ = inputs[reducer_hash]
            reduced_input = reducer(input_)
            if reduced_input is not None:
                reduced_inputs[reducer_hash] = reduced_input

        for (
            (reducer_hash, reducer_port_id, _),
            aggregator,
        ) in self._aggregators.items():
            if reducer_hash in reduced_inputs:
                aggregator.register_reduced_input(reduced_inputs[reducer_hash][reducer_port_id])

    def register_input_for_all_reducers(self, input_: Tensor) -> None:
        """
        Registers given input_ in each available statistic collection branch.

        :param input_: Tensor input to register.
        """
        self.register_inputs({hash(reducer): [input_] for reducer in self._reducers})

    def _aggregate(self) -> None:
        result = {}
        for (
            key,
            aggregator,
        ) in self._aggregators.items():
            val = aggregator.aggregate()
            result[key] = val
        return result

    def set_cache(self, statistics: TensorStatistic) -> None:
        """
        Sets cached statistics from given config and disable TensorCollector.
        :param statistics: TensorStatistic.
        """
        self._cached_statistics = statistics
        self.reset()
        self.disable()

    def create_statistics_container(self, config: Dict[str, Any]) -> TensorStatistic:
        """
        Returns a TensorStatistic instance with aggregated values.

        :param config: Aggregated values.
        :return: TensorStatistic instance.
        """
        if not self._stat_container:  # TODO(kshpv): need to remove an ability to return a Dict.
            return config
        return self._stat_container.from_config(config)

    def clear_cache(self) -> None:
        """
        Clears the cached statistics and enables TensorCollector.
        """
        self._cached_statistics = None

    def get_statistics(self) -> TensorStatistic:
        """
        Returns aggregated values in format of a TensorStatistic instance or
        a dict.

        :return: Aggregated values.
        """
        if self._cached_statistics is not None:
            return deepcopy(self._cached_statistics)

        aggregated_values = self._aggregate()
        statistics_config = {}
        for container_key, branch_key in self._stat_container_kwargs_map.items():
            statistics_config[container_key] = aggregated_values[branch_key]
        return self.create_statistics_container(statistics_config)

    def replace_aggregator(self, key: Tuple[int, int, int], aggregator: AggregatorBase) -> None:
        """
        Friend method that replaces aggregator instance on equivalent one.
        Key should be valid for for given aggregator and a statistic branch
        with key should be present in TensorCollector.

        :param key: Statistic branch key.
        :param aggregator: Aggregator instance to replace existing instance by given key.
        """
        assert key in self._aggregators
        assert key[2] == hash(aggregator)
        self._aggregators[key] = aggregator

    def reset(self):
        for aggregator in self._aggregators.values():
            aggregator.reset()

    @staticmethod
    def get_tensor_collector_inputs(
        outputs: Dict[str, Tensor], output_info: List[Tuple[int, List[str]]]
    ) -> Dict[int, List[Tensor]]:
        """
        Static method that converts all model outputs and collected output_info
        to a layout required for `register_inputs` method. This method is not a part of
        `register_inputs` to avoid all inputs passing to `TensorCollector.register_inputs` method.

        :param outputs: Target model outputs.
        :param output_info: Output info collected by a `TensorCollector.get_output_info` method.
        :returns: Model outputs in a format required by `TensorCollector.register_inputs` method.
        """
        target_inputs = {}
        for reducer, names in output_info:
            target_inputs[reducer] = [outputs[name] for name in names]
        return target_inputs


class MergedTensorCollector(TensorCollector):
    """
    Tensor collector that merge several tensor collectors in one.
    Statistics collected by a merged tensor collector automatically available
    in all tensor collectors that were merged by the merged tensor collector.
    This works because merged tensor collectors share tensor aggregators instances with
    the merged tensor collector.
    """

    def __init__(self, tensor_collectors: List[TensorCollector]) -> None:
        """
        :param tensor_collectors: Tensor collectors to merge.
        """
        super().__init__()
        aggregators: Dict[Tuple[int, int, int], List[Tuple[TensorCollector, AggregatorBase]]] = defaultdict(list)
        for tensor_collector in tensor_collectors:
            if not tensor_collector.enabled:
                continue
            self._reducers.update(tensor_collector.reducers)
            for key, aggregator in tensor_collector.aggregators.items():
                aggregators[key].append((tensor_collector, aggregator))

        for key, aggregators_to_merge in aggregators.items():
            _, unique_aggregator = aggregators_to_merge[0]
            for tensor_collector, _ in aggregators_to_merge[1:]:
                tensor_collector.replace_aggregator(key, unique_aggregator)
            self._aggregators[key] = unique_aggregator


##################################################
# Reducers
##################################################


class RawReducer(TensorReducerBase):
    def __init__(self):
        super().__init__(inplace=False)

    def get_inplace_fn(self) -> Optional[InplaceInsertionFNType]:
        return None

    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        return x


class ShapeReducer(TensorReducerBase):
    def __init__(self, inplace: bool = False):
        super().__init__(inplace=inplace)

    def _reduce_out_of_place(self, x: List[TensorType]) -> List[TensorType]:
        return [Tensor(x[0].shape)]

    def get_inplace_fn(self) -> Optional[InplaceInsertionFNType]:
        return None


class MinReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        x = x[0]
        reduction_axes = self._get_reduction_axes(x)
        return [fns.min(x, reduction_axes, keepdims=self._keepdims)]


class MaxReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        x = x[0]
        reduction_axes = self._get_reduction_axes(x)
        return [fns.max(x, reduction_axes, keepdims=self._keepdims)]


class AbsMaxReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        x = fns.abs(x[0])
        reduction_axes = self._get_reduction_axes(x)
        return [fns.max(x, reduction_axes, keepdims=self._keepdims)]


class MeanReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        x = x[0]
        reduction_axes = self._get_reduction_axes(x)
        return [fns.mean(x, reduction_axes, keepdims=self._keepdims)]


class MeanVarianceReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[TensorType]) -> List[TensorType]:
        raise NotImplementedError()


class MaxVarianceReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[TensorType]) -> List[TensorType]:
        raise NotImplementedError()


class MeanAbsMaxReducer(TensorReducerBase):
    def _reduce_out_of_place(self, x: List[TensorType]) -> List[TensorType]:
        raise NotImplementedError()


class QuantileReducerBase(TensorReducerBase):
    def __init__(
        self,
        reduction_axes: Optional[ReductionAxes] = None,
        quantile: Optional[Union[float, Tuple[float]]] = None,
        inplace: bool = False,
    ):
        super().__init__(reduction_axes=reduction_axes, inplace=False)
        self._quantile = (0.01, 0.99) if quantile is None else quantile

    def __eq__(self, __o: object) -> bool:
        return super().__eq__(__o) and self._quantile == __o._quantile

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.inplace, self._reduction_axes, tuple(self._quantile)))


class QuantileReducer(QuantileReducerBase):
    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        x = x[0]
        reduction_axes = self._get_reduction_axes(x)
        return fns.quantile(x, self._quantile, reduction_axes, keepdims=self._keepdims)


class AbsQuantileReducer(QuantileReducerBase):
    def __init__(
        self,
        reduction_axes: Optional[ReductionAxes] = None,
        quantile: Optional[Union[float, List[float]]] = None,
        inplace: bool = False,
    ):
        quantile = (0.99,) if quantile is None else quantile
        super().__init__(reduction_axes=reduction_axes, quantile=quantile, inplace=False)

    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        x = fns.abs(x[0])
        reduction_axes = self._get_reduction_axes(x)
        return fns.quantile(x, self._quantile, reduction_axes, keepdims=self._keepdims)


class BatchMeanReducer(TensorReducerBase):
    def __init__(self, inplace: bool = False):
        super().__init__(None, inplace)

    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        return [fns.mean(x[0], axis=0, keepdims=True)]


class MeanPerChReducer(TensorReducerBase):
    def __init__(self, channel_axis: int = 1, inplace: bool = False):
        super().__init__(inplace=inplace)
        self._channel_axis = channel_axis

    def _reduce_out_of_place(self, x: List[Tensor]) -> List[Tensor]:
        return [mean_per_channel(x[0], self._channel_axis)]

    def __eq__(self, __o: object) -> bool:
        return super().__eq__(__o) and self._channel_axis == __o._channel_axis

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.inplace, self._reduction_axes, self._channel_axis))


##################################################
# Aggregators
##################################################


class NoopAggregator(AggregatorBase):
    def __init__(self, num_samples: Optional[int]):
        super().__init__(None, num_samples=num_samples)

    def _register_reduced_input_impl(self, x: TensorType) -> None:
        self._container.append(x)

    def _aggregate_impl(self):
        return self._container


class ShapeAggregator(AggregatorBase):
    def __init__(self):
        super().__init__(None, num_samples=1)

    def _register_reduced_input_impl(self, x: TensorType) -> None:
        self._container = x

    def _aggregate_impl(self):
        return self._container.shape


class OnlineAggregatorBase(AggregatorBase, ABC):
    """
    Base class for aggregators which are using aggregation function fn with following property:
    fn([x1, x2, x3]) == fn([fn([x1, x2]), x3]) where x1, x2, x3 are samples to aggregate.
    Online aggregation fn([fn([x1, x2]), x3]) allows to keep memory stamp low as only
    one sample is stored during statistic collection.
    """

    def _register_reduced_input_impl(self, x: Tensor) -> None:
        online_aggregation_axes = tuple(dim - 1 for dim in self._aggregation_axes if dim != 0)
        if online_aggregation_axes:
            reduced = self._aggregation_fn(x, axis=online_aggregation_axes, keepdims=self._keepdims)
        else:
            reduced = x
        if 0 in self._aggregation_axes:
            stacked_tensors = fns.stack([reduced, *self._container], axis=0)
            aggregated = self._aggregation_fn(stacked_tensors, axis=0, keepdims=self._keepdims)
            aggregated = fns.squeeze(aggregated, 0)
            self._container = [aggregated]
        else:
            self._container.append(reduced)

    def _aggregate_impl(self) -> Tensor:
        if 0 in self._aggregation_axes and self._keepdims:
            return self._container[0]
        return fns.stack(self._container)

    @abstractmethod
    def _aggregation_fn(self, stacked_value: Tensor, axis: AggregationAxes, keepdims: bool) -> Tensor:
        pass


class MinAggregator(OnlineAggregatorBase):
    def _aggregation_fn(self, stacked_value: Tensor, axis: AggregationAxes, keepdims: bool) -> Tensor:
        return fns.min(stacked_value, axis=axis, keepdims=keepdims)


class MaxAggregator(OnlineAggregatorBase):
    def _aggregation_fn(self, stacked_value: Tensor, axis: AggregationAxes, keepdims: bool) -> Tensor:
        return fns.max(stacked_value, axis=axis, keepdims=keepdims)


class OfflineAggregatorBase(AggregatorBase, ABC):
    """
    Base class for aggregators which are using aggregation function fn which
    does not fulfill property fn([x1, x2, x3]) == fn([fn([x1, x2]), x3])
    where x1, x2, x3 are samples to aggregate. Child aggregators collect
    all samples in a container and aggregate them in one step.
    """

    def _register_reduced_input_impl(self, x: TensorType) -> None:
        self._container.append(x)

    def _aggregate_impl(self) -> Tensor:
        # Case when all registered tensors have identical shape
        if all(self._container[0].shape == x.shape for x in self._container):
            stacked_value = fns.stack(self._container)
            aggregated = self._aggregation_fn(stacked_value, axis=self._aggregation_axes, keepdims=self._keepdims)
            return fns.squeeze(aggregated, 0)
        online_axes = tuple(x - 1 for x in self._aggregation_axes if x > 0)

        # Case when some registered tensors have different shapes and
        # 0 is present in the aggregation axes
        if 0 in self._aggregation_axes:
            stacked_value, shape_after_aggregation = _move_axes_flatten_cat(self._container, online_axes)
            aggregated = self._aggregation_fn(stacked_value, axis=0, keepdims=False)
            if self._keepdims:
                aggregated = fns.reshape(aggregated, shape_after_aggregation)
            return aggregated

        # Case when some registered tensors have different shapes and
        # 0 is not present in the aggregation axes
        ret_val = []
        for tensor in self._container:
            ret_val.append(self._aggregation_fn(tensor, axis=online_axes, keepdims=self._keepdims))
        return fns.stack(ret_val, axis=0)

    @abstractmethod
    def _aggregation_fn(self, stacked_value: Tensor, axis: AggregationAxes, keepdims: bool) -> Tensor:
        pass


class MeanAggregator(OfflineAggregatorBase):
    def _aggregation_fn(self, stacked_value: Tensor, axis: AggregationAxes, keepdims: bool) -> Tensor:
        return fns.mean(stacked_value, axis=axis, keepdims=keepdims)


class MedianAggregator(OfflineAggregatorBase):
    def _aggregation_fn(self, stacked_value: Tensor, axis: AggregationAxes, keepdims: bool) -> Tensor:
        return fns.median(stacked_value, axis=axis, keepdims=keepdims)


class NoOutliersAggregatorBase(OfflineAggregatorBase, ABC):
    def __init__(
        self,
        aggregation_axes: Optional[AggregationAxes] = None,
        num_samples: Optional[int] = None,
        window_size=None,
        quantile: float = 0.01,
    ):
        super().__init__(aggregation_axes=aggregation_axes, num_samples=num_samples)
        self._window_size = window_size
        self._container = deque(maxlen=window_size)
        self._quantile = quantile

    def _aggregation_fn(self, stacked_value: Tensor, axis: int, keepdims: bool) -> Tensor:
        low_values, high_values = fns.quantile(stacked_value, q=(self._quantile, 1 - self._quantile), axis=axis)
        outliers_mask = fns.logical_or(stacked_value < low_values, high_values < stacked_value)
        aggregated = self._masked_aggregation_fn(
            stacked_samples=stacked_value,
            mask=outliers_mask,
            axis=axis,
            keepdims=keepdims,
        )
        return aggregated

    @abstractmethod
    def _masked_aggregation_fn(
        self, stacked_samples: Tensor, mask: Tensor, axis: AggregationAxes, keepdims: bool
    ) -> Tensor:
        pass

    def __eq__(self, __o: object) -> bool:
        return super().__eq__(__o) and self._quantile == __o._quantile

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self._quantile))


class MeanNoOutliersAggregator(NoOutliersAggregatorBase):
    def _masked_aggregation_fn(
        self, stacked_samples: Tensor, mask: Tensor, axis: AggregationAxes, keepdims: bool
    ) -> Tensor:
        return fns.masked_mean(stacked_samples, axis=axis, mask=mask, keepdims=keepdims)


class MedianNoOutliersAggregator(NoOutliersAggregatorBase):
    def _masked_aggregation_fn(
        self, stacked_samples: Tensor, mask: Tensor, axis: AggregationAxes, keepdims: bool
    ) -> Tensor:
        return fns.masked_median(stacked_samples, axis=axis, mask=mask, keepdims=keepdims)


class MedianAbsoluteDeviationAggregator(AggregatorBase):
    def __init__(
        self,
        aggregation_axes: Optional[AggregationAxes] = None,
        num_samples: Optional[int] = None,
        window_size=None,
    ):
        super().__init__(
            aggregation_axes=aggregation_axes,
            num_samples=num_samples,
            window_size=window_size,
        )
        if 0 not in self._aggregation_axes:
            raise NotImplementedError(
                "Aggregation without 0 dim is not supported yet for MedianAbsoluteDeviationAggregator"
            )

    def _register_reduced_input_impl(self, x: TensorType) -> None:
        return self._container.append(x)

    def _aggregate_impl(self) -> Dict[str, Tensor]:
        stacked_val, shape_after_aggregation = _move_axes_flatten_cat(
            self._container, [x - 1 for x in self._aggregation_axes if x > 0]
        )

        mask = fns.abs(stacked_val) < fns.finfo(stacked_val).eps
        median_per_ch = fns.masked_median(stacked_val, mask=mask, axis=0, keepdims=True)

        mad_values = fns.median(
            fns.abs(stacked_val - median_per_ch),
            axis=0,
            keepdims=False,
        )
        if self._keepdims:
            median_per_ch = fns.reshape(median_per_ch, shape_after_aggregation)
            mad_values = fns.reshape(mad_values, shape_after_aggregation)
        else:
            median_per_ch = fns.squeeze(median_per_ch, 0)
        return {
            MedianMADTensorStatistic.MEDIAN_VALUES_STAT: median_per_ch,
            MedianMADTensorStatistic.MAD_VALUES_STAT: mad_values,
        }


class PercentileAggregator(AggregatorBase):
    def __init__(
        self,
        percentiles_to_collect: List[float],
        aggregation_axes: Optional[AggregationAxes] = None,
        num_samples: Optional[int] = None,
        window_size=None,
    ):
        super().__init__(aggregation_axes=aggregation_axes, num_samples=num_samples)
        if 0 not in self._aggregation_axes:
            raise NotImplementedError("Aggregation without 0 dim is not supported yet for PercentileAggregator")
        self._percentiles_to_collect = percentiles_to_collect
        self._window_size = window_size
        self._container = deque(maxlen=window_size)

    def _register_reduced_input_impl(self, x: TensorType) -> None:
        return self._container.append(x)

    def _aggregate_impl(self) -> Dict[float, Tensor]:
        stacked_val, shape_after_aggregation = _move_axes_flatten_cat(
            self._container, [x - 1 for x in self._aggregation_axes if x > 0]
        )

        percentiles = fns.percentile(stacked_val, self._percentiles_to_collect, axis=0, keepdims=False)
        retval = {}
        for idx, percentile in enumerate(self._percentiles_to_collect):
            value = percentiles[idx]
            if self._keepdims:
                value = fns.reshape(value, shape_after_aggregation)
            retval[percentile] = value
        return retval


class HAWQAggregator(AggregatorBase):
    def __init__(self, num_samples: Optional[int] = None):
        super().__init__(num_samples=num_samples)
        self._container = Tensor(0.0)

    def _register_reduced_input_impl(self, x: TensorType) -> None:
        trace = fns.sum(fns.multiply(x, x))
        # NOTE: average trace?? divide by number of diagonal elements
        # TODO: revise this formula as possibly it is with an error; adopted from previous HAWQ implementation
        self._container = (self._container + trace) / x.size

    def _aggregate_impl(self) -> Tensor:
        return self._container * 2 / self._collected_samples


def _move_axes_flatten_cat(
    tensor_list: List[Tensor], aggregation_axes: Tuple[int, ...]
) -> Tuple[Tensor, Tuple[int, ...]]:
    """
    Moves aggregation axes to the beginning of the tensor shape for each tensor from the list, flattens
    and concatenates them in 0 dimension. Computes target shape for the processed tensor
    after an aggregation function is applied to it. Target shape preserves original order
    of dimensions and replaces aggregated dimensions by 1.

    :param tensor_list: Tensor list to process.
    :param aggregation_axes: Aggregation axes to move, flatten and concatenate.
    :return: Tuple of the processed tensor and
        target shape for the processed tensor after an aggregation function is applied to it.
    """
    tensor_shape = list(tensor_list[0].shape)

    # Transpose dims to move aggregation axes forward
    transpose_dims = list(range(len(tensor_shape)))
    for idx, axis in enumerate(aggregation_axes):
        transpose_dims[axis], transpose_dims[idx] = transpose_dims[idx], transpose_dims[axis]

    # Shape to flatten aggregation axes
    reshape_shape = [-1] + [tensor_shape[dim] for dim in transpose_dims][len(aggregation_axes) :]

    reshaped_tensors = []
    for tensor in tensor_list:
        transposed_t = fns.transpose(tensor, transpose_dims)
        reshaped_tensors.append(fns.reshape(transposed_t, reshape_shape))

    shape_after_aggregation = tuple(1 if idx in aggregation_axes else dim for idx, dim in enumerate(tensor_shape))
    return fns.concatenate(reshaped_tensors, axis=0), shape_after_aggregation


REDUCERS_MAP = {
    StatisticsType.MIN: MinReducer,
    StatisticsType.MAX: MaxReducer,
    StatisticsType.ABS_MAX: AbsMaxReducer,
    StatisticsType.MEAN: MeanReducer,
    StatisticsType.QUANTILE: QuantileReducer,
    StatisticsType.ABS_QUANTILE: AbsQuantileReducer,
}

AGGREGATORS_MAP = {
    AggregatorType.MIN: MinAggregator,
    AggregatorType.MAX: MaxAggregator,
    AggregatorType.MEAN: MeanAggregator,
    AggregatorType.MEAN_NO_OUTLIERS: MeanNoOutliersAggregator,
    AggregatorType.MEDIAN: MedianAggregator,
    AggregatorType.MEDIAN_NO_OUTLIERS: MedianNoOutliersAggregator,
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import deque

import onnx

from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.onnx.graph.metatypes.onnx_metatypes import ONNXDequantizeLinearMetatype
from nncf.onnx.graph.metatypes.onnx_metatypes import ONNXQuantizeLinearMetatype
from nncf.onnx.graph.transformations.commands import ONNXQDQNodeRemovingCommand
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint


def remove_fq_from_inputs(model: onnx.ModelProto, nncf_graph: NNCFGraph) -> onnx.ModelProto:
    """
    This method removes the activation Quantizer nodes from the model.
    It's needed for the further bias shift calculation that relates on quantized weights.

    :param model: onnx.ModelProto instance.
    :param nncf_graph: NNCFGraph instance.
    :return: onnx.ModelProto instance without activation Quantizer nodes.
    """
    transformation_layout = TransformationLayout()
    model_transformer = ModelTransformerFactory.create(model)

    seen_nodes = []
    nodes_queue = deque(nncf_graph.get_input_nodes())
    while nodes_queue:
        current_node = nodes_queue.popleft()
        current_node_name = current_node.node_name

        if current_node_name in seen_nodes:
            continue

        seen_nodes.append(current_node_name)
        if current_node.metatype in [ONNXQuantizeLinearMetatype, ONNXDequantizeLinearMetatype]:
            target_point = ONNXTargetPoint(TargetType.LAYER, current_node_name, 0)
            command = ONNXQDQNodeRemovingCommand(target_point)
            transformation_layout.register(command)
        nodes_queue.extend(nncf_graph.get_next_nodes(current_node))

    return model_transformer.transform(transformation_layout)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import inspect
from typing import Any, Callable, List, Optional, TypeVar, Union

from nncf.telemetry.events import MODEL_BASED_CATEGORY
from nncf.telemetry.events import get_current_category
from nncf.telemetry.events import get_model_based_category
from nncf.telemetry.events import telemetry_category
from nncf.telemetry.extractors import CollectedEvent
from nncf.telemetry.extractors import TelemetryExtractor
from nncf.telemetry.extractors import VerbatimTelemetryExtractor
from nncf.telemetry.wrapper import telemetry

TFunction = TypeVar("TFunction", bound=Callable[..., Any])


class tracked_function:
    """
    Decorator to add telemetry events to a given function execution. The call to a decorated function
    will in general result in a telemetry session being created and a set of events being sent before
    function execution. The category of the session and events will be determined by parameters to the decorator.
    """

    def __init__(self, category: str = None, extractors: Optional[List[Union[str, TelemetryExtractor]]] = None) -> None:
        """
        :param category: A category to be attributed to the events. If set to None, no events will be sent.
        :param extractors: Add argument names in this list as string values to send an event with an "action" equal to
            the argument name and "label" equal to the argument value before the function start.
            The argument must be either a string or a dictionary of strings. If that is not the case, instead of
            argument names you can specify an object of a customized `TelemetryExtractor` subclass; use the same
            approach for more complex event reporting.
        """
        self._category = category
        if extractors is not None:
            self._collectors = [VerbatimTelemetryExtractor(x) if isinstance(x, str) else x for x in extractors]
        else:
            self._collectors = []

    def __call__(self, fn: TFunction) -> TFunction:
        fn_signature = inspect.signature(fn)

        @functools.wraps(fn)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            bound_args = fn_signature.bind(*args, **kwargs)
            bound_args.apply_defaults()

            if self._category == MODEL_BASED_CATEGORY:
                model_argument = bound_args.arguments.get("model", None)
                self._category = get_model_based_category(model_argument)

            events: List[CollectedEvent] = []
            for collector in self._collectors:
                argname = collector.argname
                argvalue = bound_args.arguments[argname] if argname else None
                event = collector.extract(argvalue)
                events.append(event)

            previous_category = get_current_category()
            with telemetry_category(self._category) as category:
                if category is not None:
                    if category != previous_category:
                        telemetry.start_session(self._category)
                    for event in events:
                        telemetry.send_event(
                            event_category=category,
                            event_action=event.name,
                            event_label=event.data,
                            event_value=event.int_data,
                        )

                retval = fn(*args, **kwargs)

                if category is not None and category != previous_category:
                    telemetry.end_session(self._category)
            return retval

        return wrapped  # type: ignore[return-value]



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bisect import bisect_right
from typing import Any, Dict, List, Optional

import numpy as np

from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage


class PolynomialDecaySchedule:
    """
    This schedule applies a polynomial decay function to an epoch index.
    For more details about polynomial decay see the [paper](https://arxiv.org/abs/1710.01878).
    """

    def __init__(self, initial_value: float, target_value: float, target_epoch: int, power: float, concave: bool):
        """
         Initializes a schedule with a polynomial decay function.

        :param initial_value: The initial value at which the schedule begins.
        :param target_value: The final value at which the schedule ends.
        :param target_epoch: Zero-based index of the epoch from which
            the function value will be equal to the `target_value` value.
        :param power: Exponent to be used in the polynomial decay function.
        :param concave: If true, then the `target_value` will be approached in a
            concave manner, and in a convex manner otherwise.
        """
        self.initial_value = initial_value
        self.target_value = target_value
        self.target_epoch = target_epoch
        self.power = power
        self.concave = concave

    def __call__(self, epoch: int, step: Optional[int] = None, steps_per_epoch: Optional[int] = None) -> float:
        """
        Calculates the value of the polynomial decay function. Two ways are possible:
            - Using epoch index only.
            - Using epoch index and step index. For this case, `steps_per_epoch` should
                be provided too.

        :param epoch: Zero-based epoch index for which the function value should be got.
        :param step: Local step index in `epoch` i.e. `step` should be in interval [0, steps_per_epoch).
            Used only if `steps_per_epoch` was provided.
        :param steps_per_epoch: A number of steps per epoch.
        """
        if self.target_epoch == 0:
            return self.target_value

        if step is not None and steps_per_epoch is not None:
            fractional_epoch = epoch + step / steps_per_epoch
            progress = fractional_epoch / self.target_epoch
        else:
            progress = epoch / self.target_epoch
        progress = min(1.0, max(0.0, progress))

        if self.concave:
            value = self.target_value - (self.target_value - self.initial_value) * np.power(1 - progress, self.power)
        else:
            value = self.initial_value + (self.target_value - self.initial_value) * np.power(progress, self.power)

        return float(value)


class MultiStepSchedule:
    """
    This schedule applies a piecewise constant function to an epoch index
    """

    def __init__(self, boundaries: List[int], values: List[float]):
        """
        Initializes a schedule with a piecewise constant function.

        :param boundaries: List of zero-based epoch indices. Must be increasing.
        :param values: List of floats that specifies the values for the intervals
            defined by `boundaries`. It should have one more element than `boundaries`.
        :raises ValueError: If the number of elements in the `values` list does not
            equal to the number of elements in the `boundaries` list plus one.
        """
        if len(boundaries) + 1 != len(values):
            raise ValueError("The length of `values` should be 1 more than the length of `boundaries`")

        self.boundaries = boundaries
        self.values = values

    def __call__(self, epoch: int) -> float:
        """
        Calculates the value of the piecewise constant function for a given epoch index.
        The output of this call is `values[0]` when `epoch` < `boundaries[0]`,
        `values[1]` when  `boundaries[0]` <= `epoch` < `boundaries[1]`, ... , and
        `values[-1]` when `epoch` >= boundaries[-1].

        :param epoch: Zero-based epoch index for which the function value should be got.
        :return: The value of the piecewise constant function for a given epoch index.
        """
        pos = bisect_right(self.boundaries, epoch)
        return self.values[pos]


class ExponentialDecaySchedule:
    """
    This schedule applies an exponential decay function to an epoch index,
    considering a provided `initial_value` value. It is computed as:

        current_value = initial_value * decay_rate ^ (epoch / target_epoch),

    where `decay_rate` is equal to target_value / initial_value.
    """

    def __init__(self, initial_value: float, target_value: float, target_epoch: int):
        """
        Initializes a schedule with an exponential decay function.

        :param initial_value: The initial value at which the schedule begins.
        :param target_value: The final value at which the schedule end.
        :param target_epoch: Zero-based index of the epoch from which
            the function value will be equal to the `target_value` value.
        """
        self.initial_value = initial_value
        self.target_value = target_value
        self.target_epoch = target_epoch
        self.decay_rate = target_value / initial_value

    def __call__(self, epoch: int) -> float:
        """
        Calculates the value of the exponential decay function for a given epoch index.

        :param epoch: Zero-based epoch index for which the function value should be got.
        :return: The value of the exponential decay function for a given epoch index.
        """
        if self.target_epoch == 0:
            return self.target_value

        value = self.initial_value * float(np.power(self.decay_rate, epoch / self.target_epoch))
        return max(value, self.target_value)


class BaseCompressionScheduler(CompressionScheduler):
    """
    Contains the implementation of the basic functionality of the scheduler.

    The `step()` and `epoch_step()` methods of the compression scheduler must be
    called at the beginning of each training step and epoch, respectively.

    ```
    for epoch in range(0, num_epochs):
        scheduler.epoch_step()
        for i, (x, y) in enumerate(dataset):
             scheduler.step()
             ...
    ```
    """

    def __init__(self) -> None:
        """
        Initializes the internal state of the compression scheduler specified by:
            - `current_step` is the index of the global training step, counted
            from 0 to the end of training. The initial value is -1
            - `current_epoch` is the training epoch index (numbering from zero).
            The initial value is -1.

        The `current_step` and `current_epoch` specify the training step and epoch,
        respectively, for which the compression scheduler has updated the state of
        the compression method, in particular its hyperparameters. It means that
        the compression method is configured and ready to continue training at
        `current_step` and `current_epoch`.

        When `current_step` is -1, it means that the compression scheduler did not
        update the compression method state taking into account the training step,
        there is the same for current_epoch is -1.
        """
        self._current_step = -1
        self._current_epoch = -1

    @property
    def current_step(self) -> int:
        """
        Return current step.

        :return: Current step.
        """
        return self._current_step

    @property
    def current_epoch(self) -> int:
        """
        Return current epoch.

        :return: Current epoch.
        """
        return self._current_epoch

    def step(self, next_step: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training step to prepare
        the compression method to continue training the model in the `next_step`.

        :param next_step: The global step index for which the compression scheduler
            will update the state of the compression method.
        """
        if next_step is None:
            next_step = self._current_step + 1
        self._current_step = next_step

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training epoch to prepare
        the compression method to continue training the model in the `next_epoch`.

        :param next_epoch: The epoch index for which the compression scheduler
            will update the state of the compression method.
        """
        if next_epoch is None:
            next_epoch = self._current_epoch + 1
        self._current_epoch = next_epoch

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Loads the compression scheduler state, but does not update the state of the
        compression method.

        :param state: Output of `get_state()` method.
        """
        self._current_step = state["current_step"]
        self._current_epoch = state["current_epoch"]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression scheduler state.

        :return: The compression scheduler state.
        """
        return {"current_step": self._current_step, "current_epoch": self._current_epoch}


class StubCompressionScheduler(CompressionScheduler):
    def step(self, next_step: Optional[int] = None) -> None:
        pass

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        pass

    def load_state(self, state: Dict[str, Any]) -> None:
        pass

    def get_state(self) -> Dict[str, Any]:
        return {}

    def compression_stage(self) -> CompressionStage:
        return CompressionStage.FULLY_COMPRESSED



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import weakref
from collections import OrderedDict
from collections import defaultdict
from collections import deque
from contextlib import contextmanager
from typing import Callable, DefaultDict, Dict, List, Optional, Union

import torch

from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.hook_handle import HookHandle
from nncf.common.hook_handle import add_op_to_registry
from nncf.common.utils.api_marker import api
from nncf.common.utils.debug import is_debug
from nncf.common.utils.patcher import PATCHER
from nncf.torch.dynamic_graph.graph import DynamicGraph
from nncf.torch.dynamic_graph.graph import DynamicGraphNode
from nncf.torch.dynamic_graph.graph import DynamicGraphNodeParameters
from nncf.torch.dynamic_graph.graph import TensorMetaComparator
from nncf.torch.dynamic_graph.op_input_processing import OperatorInput
from nncf.torch.dynamic_graph.operation_address import OperationAddress
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.dynamic_graph.scope import ScopeElement
from nncf.torch.dynamic_graph.trace_tensor import TensorMeta
from nncf.torch.dynamic_graph.trace_tensor import TracedTensorMixin


class ThreadLocalGlobalContext(threading.local):
    def __init__(self):
        super().__init__()
        self.context = None


_CURRENT_CONTEXT = ThreadLocalGlobalContext()


class PreHookId:
    def __init__(self, op_address: OperationAddress, input_port_id: int):
        self.op_address = op_address
        self.input_port_id = input_port_id

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __str__(self):
        return str(self.op_address) + "|INPUT{}".format(self.input_port_id)

    def __hash__(self):
        return hash(str(self))


class TracingThreadLocals(threading.local):
    def __init__(self):
        super().__init__()
        self.reset()

    def reset(self):
        self.scopes = []
        self.module_call_stack = []
        self.in_operator = False
        self.in_parameter_trace = False
        self.num_nested_hooks = 0
        self.base_module_replica = None
        self.operator_counters = {}
        self.node_call_tracker = {}
        self.traced_tensor_weakrefs = []
        self.nested_contexts_stack = []
        self.processed_parameters = {}


class CopySafeThreadingVars:
    """A class holding variables that are related to threading and
    thus impossible to deepcopy. The deepcopy will simply return a
    new object without copying, but won't fail."""

    def __init__(self):
        self.thread_local = TracingThreadLocals()
        self.cond = threading.Condition()

    def __deepcopy__(self, memo):
        return CopySafeThreadingVars()


class TracingContext:
    def __init__(self):
        self.graph = DynamicGraph()

        self._post_hooks: DefaultDict[OperationAddress, Dict[str, Callable]] = defaultdict(OrderedDict)
        self._pre_hooks: DefaultDict[PreHookId, Dict[str, Callable]] = defaultdict(OrderedDict)
        self._num_nested_hooks = 0
        self.reused_parameters = []

        self._threading = CopySafeThreadingVars()

        self._n_instances_searching_graph = 0

        self._is_tracing = True
        self._is_forwarding = False
        self._may_add_nodes = True
        self._input_comparators_per_scope = []
        self.global_buffer_store = {}

        self._trace_dynamic_graph = False

        # ELASTIC DEPTH PARAMS
        self.elastic_depth = False
        self.skipped_blocks = []
        self.in_skipped_block = False
        self.start_node_name_of_skipped_block = []
        self.end_node_name_of_skipped_block = []
        self.active_block_indexes = None
        self.tensor_cache = None
        self._ordinals_ids = None

    def __enter__(self):
        # For DataParallel, this relies on having the same compressed context for
        # all replicas. Otherwise we will have data races on setting and reading the global _CURRENT_CONTEXT
        # variable, which will in turn lead to DP-specific runtime errors such as
        # "'_thread._local' object has no attribute 'scopes'"
        self._threading.thread_local.nested_contexts_stack.append(get_current_context())
        set_current_context(self)

        return self

    def __exit__(self, *args):
        previous_context = self._threading.thread_local.nested_contexts_stack.pop(-1)
        for traced_tensor_weakref in self._threading.thread_local.traced_tensor_weakrefs:
            tt = traced_tensor_weakref()
            if tt is None or not isinstance(tt, TracedTensorMixin):
                continue
            if previous_context is None:
                tt.strip()
            elif previous_context is not self:
                previous_context.register_traced_tensor(tt)

        if previous_context is not self:
            self._reset_thread_local()

            if is_debug():
                self.reset_node_call_counters()

        set_current_context(previous_context)

    def find_operator_node(
        self, tensor_metas: List[Optional[TensorMeta]], op_address: OperationAddress
    ) -> Optional[DynamicGraphNode]:
        with self._threading.cond:
            self._n_instances_searching_graph += 1

        node = self.graph.find_node(op_address, tensor_metas, self._input_comparators_per_scope)

        with self._threading.cond:
            self._n_instances_searching_graph -= 1
            self._threading.cond.notify_all()
        return node

    def register_global_buffer(self, name: str, buffer):
        self.global_buffer_store[name] = buffer

    def register_traced_tensor(self, tt: torch.Tensor):
        """
        Registers a weak reference to a traced tensor in the context so that in case
        the block under context retains a reference to an intermediate tensor somewhere,
        the context strips this traced tensor upon context exit.

        :param tt: A tensor with TracedTensorMixin tracing capabilities to be registered.
        """
        wr = weakref.ref(tt)
        self._threading.thread_local.traced_tensor_weakrefs.append(wr)

    def register_processed_parameter(self, param_name: str, tensor: torch.Tensor) -> None:
        """
        Registers the processed parameter in the context to avoid double calculation of hooks
        for the same parameters.

        :param param_name: The parameter name.
        :param tensor: The processed parameter.
        """
        self._threading.thread_local.processed_parameters[param_name] = tensor

    def get_processed_parameter(self, param_name: str) -> Union[torch.Tensor, None]:
        """
        Return the processed parameter by name.

        :param param_name: The parameter name.
        :return: The processed parameter by name if found, otherwise None.
        """
        return self._threading.thread_local.processed_parameters.get(param_name, None)

    def maybe_add_node(
        self,
        inputs: OperatorInput,
        tensor_metas: List[Optional[TensorMeta]],
        op_address: OperationAddress,
        module_attrs: BaseLayerAttributes = None,
        ignored_algorithms: List[str] = None,
        is_called_inside_nncf_module: bool = False,
    ) -> Optional[DynamicGraphNode]:
        if not self._may_add_nodes:
            return None
        with self._threading.cond:
            while self._n_instances_searching_graph > 0:
                self._threading.cond.wait()
            # Another thread may have added a node inside this block,
            # so we need to check again if a node is already added.
            node = self.graph.find_node(op_address, tensor_metas, self._input_comparators_per_scope)
            if node is None:
                mid = id(self.get_current_module())
                node = self.graph.add_node(
                    op_address,
                    tensor_metas,
                    self._input_comparators_per_scope,
                    inputs,
                    DynamicGraphNodeParameters(
                        module_attrs, ignored_algorithms, is_called_inside_nncf_module, calling_module_id=mid
                    ),
                )
        return node

    def get_caller_context(self, operator_name: str) -> OperationAddress:
        """
        Designed to work in the following way - for each scope the context will track the number of the calls to the
        operators with the name operator_name (call_order). The counter values are preserved until reset by a
        corresponding member function of the context, which must be called after each model iteration - this is
        usually handled inside NNCF. This mechanism allows to discern between multiple function calls inside the same
        module that would each require their own instance of compression layers - for instance, multiple `relu`
        function calls (either on their own or inside a `for` cycle), and at the same moment allow the checkpoints to
        be loaded if the model had changed in the meantime in a way that does not impact the major function call
        order (e.g. if comments were added to the .py file with the model)
        """

        call_order = self.get_operator_call_count_in_scope(operator_name, self.scope)

        op_address = OperationAddress(operator_name, self.scope, call_order)
        return op_address

    def reset_scope_operator_call_counters(self):
        """
        Must be called after each "forward" operation of the model that is made
        within this context
        """
        self._threading.thread_local.operator_counters = {}

    @staticmethod
    def _get_operator_counter_key(operator_name: str, scope: Scope):
        return "{}_{}".format(str(scope), operator_name)

    def register_operator_call(self, operator_name: str, scope: Scope):
        key = self._get_operator_counter_key(operator_name, scope)
        if key in self._threading.thread_local.operator_counters:
            self._threading.thread_local.operator_counters[key] += 1
        else:
            self._threading.thread_local.operator_counters[key] = 1

    def get_operator_call_count_in_scope(self, operator_name: str, scope: Scope):
        key = self._get_operator_counter_key(operator_name, scope)
        if key in self._threading.thread_local.operator_counters:
            return self._threading.thread_local.operator_counters[key]
        return 0

    def reset_operator_call_count_in_scope(self, scope):
        scoped_op_name = str(scope)
        for key in self._threading.thread_local.operator_counters:
            if scoped_op_name in key:
                self._threading.thread_local.operator_counters[key] = 0

    def push_scope(self, called_module: torch.nn.Module):
        relative_scopes_list = self._get_scope_relative_to_last_registered_module_call(called_module)
        self.module_call_stack.append(called_module)
        self.relative_scopes_stack.append(relative_scopes_list)

    def pop_scope(self):
        self.relative_scopes_stack.pop()
        self.module_call_stack.pop()

    def register_pre_hook(self, fn: Callable, op_address: OperationAddress, input_port_id: int) -> HookHandle:
        pre_hook_id = PreHookId(op_address, input_port_id)
        return add_op_to_registry(self._pre_hooks[pre_hook_id], fn)

    def execute_pre_hooks(self, op_address: OperationAddress, op_inputs: OperatorInput) -> OperatorInput:
        in_op = getattr(self, "in_operator", False)
        self.in_operator = False
        self._threading.thread_local.num_nested_hooks += 1

        pre_hook_ids_for_curr_op = [x for x in self._pre_hooks if x.op_address == op_address]
        pre_hook_ids_for_curr_op = sorted(pre_hook_ids_for_curr_op, key=lambda x: x.input_port_id)
        for pre_hook_id in pre_hook_ids_for_curr_op:
            hook_list_for_current_input_port = self._pre_hooks[pre_hook_id].values()
            input_arg_to_process = pre_hook_id.input_port_id
            for hook in hook_list_for_current_input_port:
                op_inputs[input_arg_to_process] = hook(op_inputs[input_arg_to_process])
        self._threading.thread_local.num_nested_hooks -= 1
        self.in_operator = in_op
        return op_inputs

    def register_post_hook(self, fn: Callable, op_address: OperationAddress) -> HookHandle:
        return add_op_to_registry(self._post_hooks[op_address], fn)

    def execute_post_hooks(self, op_address: OperationAddress, outputs):
        in_op = getattr(self, "in_operator", False)
        self.in_operator = False
        self._threading.thread_local.num_nested_hooks += 1
        if op_address in self._post_hooks:
            for hook in self._post_hooks[op_address].values():
                outputs = hook(outputs)
        self._threading.thread_local.num_nested_hooks -= 1
        self.in_operator = in_op
        return outputs

    @property
    def is_tracing(self) -> bool:
        return self._is_tracing

    def disable_tracing(self):
        self._is_tracing = False

    def enable_tracing(self):
        self._is_tracing = True

    @property
    def is_forwarding(self) -> bool:
        return self._is_forwarding

    def disable_forwarding(self):
        self._is_forwarding = False

    def enable_forwarding(self):
        self._is_forwarding = True

    def enable_node_additions(self):
        self._may_add_nodes = True

    def disable_node_additions(self):
        self._may_add_nodes = False

    def add_node_comparators(self, scopes_to_apply: List[str], node_input_comparator: TensorMetaComparator = None):
        self._input_comparators_per_scope.append((node_input_comparator, scopes_to_apply))

    @property
    def base_module_thread_local_replica(self) -> torch.nn.Module:
        return self._threading.thread_local.base_module_replica

    @base_module_thread_local_replica.setter
    def base_module_thread_local_replica(self, value: torch.nn.Module):
        self._threading.thread_local.base_module_replica = value

    @property
    def in_operator(self):
        return self._threading.thread_local.in_operator

    @in_operator.setter
    def in_operator(self, val):
        self._threading.thread_local.in_operator = val

    @property
    def in_parameter_trace(self):
        return self._threading.thread_local.in_parameter_trace

    @in_parameter_trace.setter
    def in_parameter_trace(self, val):
        self._threading.thread_local.in_parameter_trace = val

    @property
    def module_call_stack(self) -> List[torch.nn.Module]:
        return self._threading.thread_local.module_call_stack

    def get_current_module(self) -> Optional[torch.nn.Module]:
        if self.module_call_stack:
            return self.module_call_stack[-1]
        return None

    @property
    def relative_scopes_stack(self) -> List[Scope]:
        return self._threading.thread_local.scopes

    @property
    def trace_dynamic_graph(self) -> bool:
        return self._trace_dynamic_graph

    def disable_trace_dynamic_graph(self):
        if not self.graph.is_graph_with_iteration_modules():
            self._trace_dynamic_graph = False

    def enable_trace_dynamic_graph(self):
        self._trace_dynamic_graph = True

    def _reset_thread_local(self):
        self._threading.thread_local.reset()

    def register_node_call(self, node: DynamicGraphNode):
        if node.node_id in self._threading.thread_local.node_call_tracker:
            self._threading.thread_local.node_call_tracker[node.node_id] += 1
        else:
            self._threading.thread_local.node_call_tracker[node.node_id] = 1

    def reset_node_call_counters(self):
        for k, _ in self._threading.thread_local.node_call_tracker.items():
            self._threading.thread_local.node_call_tracker[k] = 0

    def get_node_call_counter_dict(self):
        return self._threading.thread_local.node_call_tracker

    def _get_scope_relative_to_last_registered_module_call(self, module) -> Scope:
        module_class = module.__class__.__name__
        if not self.module_call_stack:
            return Scope(
                [
                    ScopeElement(module_class),
                ]
            )
        q = deque([(tuple(), self.module_call_stack[-1])])
        while q:
            scope_parts, top = q.popleft()
            if module is top:
                return Scope(list(scope_parts))
            for name, child in top.named_children():
                scope_element = ScopeElement(child.__class__.__name__, name)
                q.append((scope_parts + (scope_element,), child))
        return Scope(
            [
                ScopeElement(module_class),
            ]
        )

    @property
    def scope(self) -> Scope:
        stack_copy = self.relative_scopes_stack.copy()
        scope_el_list = []
        for relative_scope in stack_copy:
            for scope_element in relative_scope.scope_elements:
                scope_el_list.append(scope_element)
        return Scope(scope_el_list)

    def reset_graph(self):
        self.graph = DynamicGraph()

    def set_active_skipped_block(self, block_indexes: List[int]):
        if self.active_block_indexes is not None:
            self.start_node_name_of_skipped_block = []
            self.end_node_name_of_skipped_block = []
        self.active_block_indexes = block_indexes
        if self.elastic_depth and len(block_indexes) > 0:
            for block_index in block_indexes:
                self.start_node_name_of_skipped_block.append(self.skipped_blocks[block_index].start_node_name)
                self.end_node_name_of_skipped_block.append(self.skipped_blocks[block_index].end_node_name)

    def set_elastic_blocks(self, blocks: List["BuildingBlock"] = None):  # noqa: F821
        if blocks is not None and isinstance(blocks, list):
            if len(blocks) == 0:
                self.skipped_blocks = []
            elif isinstance(blocks[0], str):
                self.skipped_blocks = [blocks]
            else:
                self.skipped_blocks = blocks


def set_current_context(c: TracingContext):
    _CURRENT_CONTEXT.context = c


@api(canonical_alias="nncf.torch.no_nncf_trace")
@contextmanager
def no_nncf_trace():
    ctx = get_current_context()
    if ctx is not None and ctx.is_tracing:
        ctx.disable_tracing()
        yield
        ctx.enable_tracing()
    else:
        yield


@api(canonical_alias="nncf.torch.forward_nncf_trace")
@contextmanager
def forward_nncf_trace():
    ctx = get_current_context()
    if ctx is not None and not ctx.is_forwarding:
        ctx.enable_forwarding()
        yield
        ctx.disable_forwarding()
    else:
        yield


def get_current_context() -> TracingContext:
    return _CURRENT_CONTEXT.context


@api(canonical_alias="nncf.torch.disable_tracing")
def disable_tracing(method):
    """
    Patch a method so that it will be executed within no_nncf_trace context
    :param method: A method to patch.
    """

    def no_nncf_trace_wrapper(self, fn, *args, **kwargs):
        with no_nncf_trace():
            return fn(*args, **kwargs)

    PATCHER.patch(method, no_nncf_trace_wrapper)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import Deque, List, Type, TypeVar

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import OperatorMetatype

TModel = TypeVar("TModel")


def transform_to_inference_graph(
    nncf_graph: NNCFGraph,
    input_nodes: List[NNCFNode],
    shapeof_metatypes: List[Type[OperatorMetatype]],
    dropout_metatypes: List[Type[OperatorMetatype]],
    preserved_metatypes: List[Type[OperatorMetatype]],
) -> NNCFGraph:
    """
    This method contains inplace pipeline of the passes that uses to provide inference graph without constant flows.

    :param nncf_graph: NNCFGraph instance for the transformation.
    :param input_nodes: List of input nodes for the given NNCFGraph.
    :param shapeof_metatypes: List of backend-specific ShapeOf metatypes.
    :param dropout_metatypes: List of backend-specific Dropout metatypes.
    :return: NNCFGraph in the inference style.
    """
    shapeof_subgraphs = find_shapeof_subgraphs(nncf_graph, shapeof_metatypes, input_nodes)
    preserved_nodes = find_preserved_nodes(nncf_graph, shapeof_subgraphs, preserved_metatypes)
    constant_subgraphs = find_constant_subgraphs(nncf_graph, input_nodes)

    nodes_to_drop = set([*shapeof_subgraphs, *constant_subgraphs]).difference(preserved_nodes)
    nncf_graph.remove_nodes_from(nodes_to_drop)

    remove_nodes_and_reconnect_graph(nncf_graph, dropout_metatypes)
    return nncf_graph


def find_shapeof_subgraphs(
    nncf_graph: NNCFGraph,
    shapeof_metatypes: List[Type[OperatorMetatype]],
    input_nodes: List[NNCFNode],
) -> List[NNCFNode]:
    """
    Returns a list of nodes belonging to ShapeOf subgraphs.

    :param nncf_graph: The input graph to be analyzed.
    :param shapeof_metatypes: A list of metatypes representing backend-specific
        ShapeOf operations.
    :param input_nodes: A list of nodes designated as graph inputs. These nodes are
        used to identify which nodes depend on input data.
    :return: A list of nodes belonging to ShapeOf subgraphs.
    """
    shapeof_subgraphs = set()
    shape_of_nodes = []
    infer_nodes = []

    nodes_queue = collections.deque(input_nodes)
    while nodes_queue:
        node = nodes_queue.pop()
        if node.metatype in shapeof_metatypes:
            shape_of_nodes.append(node)
            continue
        if node.node_name in infer_nodes:
            continue
        infer_nodes.append(node.node_name)
        nodes_queue.extend(nncf_graph.get_next_nodes(node))

    for shape_of_node in shape_of_nodes:
        shapeof_subgraphs.add(shape_of_node)

        shape_of_queue: Deque[NNCFNode] = collections.deque()
        shape_of_queue.extend(nncf_graph.get_next_nodes(shape_of_node))
        while shape_of_queue:
            node = shape_of_queue.pop()
            if node in shapeof_subgraphs or node.node_name in infer_nodes:
                continue
            shapeof_subgraphs.add(node)
            # traverse forward and backward to exclude full shape of subgraph
            # recursion excluded due to infer_nodes list around subgraph shape
            shape_of_queue.extend(nncf_graph.get_next_nodes(node) + nncf_graph.get_previous_nodes(node))

    return list(shapeof_subgraphs)


def find_preserved_nodes(
    graph: NNCFGraph,
    shapeof_subgraphs: List[NNCFNode],
    preserved_metatypes: List[Type[OperatorMetatype]],
) -> List[NNCFNode]:
    """
    :param graph: The input graph to be analyzed.
    :param shapeof_subgraphs: A list of nodes belonging to ShapeOf subgraphs.
    :param preserved_metatypes: Backend-specific metatypes that require preserving
        float subgraphs when removing the ShapeOf subgraph.
    :return: A list of nodes in float subgraphs of ShapeOf subgraphs.
    """
    preserved_nodes = set()
    for node in graph.get_nodes_by_metatypes(preserved_metatypes):
        for e in graph.get_input_edges(node):
            if e.from_node in shapeof_subgraphs and e.dtype == Dtype.FLOAT:
                preserved_nodes.add(e.from_node)

    queue = collections.deque(preserved_nodes)
    while queue:
        node = queue.pop()

        for e in graph.get_input_edges(node):
            if e.from_node in preserved_nodes:
                continue

            if e.dtype == Dtype.FLOAT and e.from_node in shapeof_subgraphs:
                queue.append(e.from_node)
                preserved_nodes.add(e.from_node)

    return list(preserved_nodes)


def remove_nodes_and_reconnect_graph(
    nncf_graph: NNCFGraph,
    metatypes: List[Type[OperatorMetatype]],
) -> NNCFGraph:
    """
    Removes nodes with metatypes specified by `metatypes` parameter from
    the provided NNCFGraph instance and connects previous node of a matched node
    with next nodes of a matched node inplace for each matched node.
    Matched nodes should have only one input node and only one output port.

    :param nncf_graph: NNCFGraph instance for the transformation.
    :param metatypes: List of backend-specific metatypes.
    :return: Resulting NNCFGraph.
    """
    if not metatypes:
        return nncf_graph

    nodes_to_drop = []
    for node in nncf_graph.get_nodes_by_metatypes(metatypes):
        if node.metatype in metatypes:
            nodes_to_drop.append(node)

            prev_nodes = nncf_graph.get_previous_nodes(node)
            input_edges = nncf_graph.get_input_edges(node)
            assert len(prev_nodes) == len(input_edges) == 1
            prev_node = prev_nodes[0]
            input_edge = input_edges[0]
            assert not input_edge.parallel_input_port_ids

            # nncf_graph.get_next_edges is not used to preserve
            # parallel_input_port_ids
            for output_node in nncf_graph.get_next_nodes(node):
                output_edge = nncf_graph.get_edge(node, output_node)
                # Connects previous node with all next nodes
                # to keep NNCFGraph connected.
                assert input_edge.dtype == output_edge.dtype
                assert input_edge.tensor_shape == output_edge.tensor_shape
                nncf_graph.add_edge_between_nncf_nodes(
                    from_node_id=prev_node.node_id,
                    to_node_id=output_edge.to_node.node_id,
                    tensor_shape=input_edge.tensor_shape,
                    input_port_id=output_edge.input_port_id,
                    output_port_id=input_edge.output_port_id,
                    dtype=input_edge.dtype,
                    parallel_input_port_ids=output_edge.parallel_input_port_ids,
                )
    nncf_graph.remove_nodes_from(nodes_to_drop)
    return nncf_graph


def find_constant_subgraphs(
    nncf_graph: NNCFGraph,
    input_nodes: List[NNCFNode],
) -> List[NNCFNode]:
    """
    Returns a list of nodes belonging to constant subgraphs.

    :param nncf_graph: The input graph to be analyzed.
    :param input_nodes: A list of nodes designated as graph inputs. These nodes are
        used to identify which nodes depend on input data.
    :return: A list of nodes belonging to constant subgraphs.
    """
    if not input_nodes:
        return []

    visited_nodes = set()
    nodes_queue = collections.deque(input_nodes)
    while nodes_queue:
        node = nodes_queue.pop()
        if node in visited_nodes:
            continue
        visited_nodes.add(node)
        nodes_queue.extend(nncf_graph.get_next_nodes(node))
    constant_nodes = [node for node in nncf_graph.get_all_nodes() if node not in visited_nodes]

    return constant_nodes



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ValidationError(Exception):
    """
    Raised when an internal NNCF validation check fails,
    for example, if the user supplied an invalid or inconsistent set of arguments.
    """

    pass


class InternalError(Exception):
    """
    Raised when an internal error occurs within the NNCF framework.

    This exception is raised when an unexpected internal error occurs during the execution
    of NNCF. It indicates a situation where the code encountered an unexpected condition.

    """

    pass


class UnsupportedDatasetError(Exception):
    """
    Raised when an unsupported dataset is encountered
    """

    pass


class InvalidPathError(Exception):
    """
    Raised when an invalid path is provided
    """

    pass


class UnsupportedBackendError(Exception):
    """
    Raised when an unsupported backend is specified
    """

    pass


class InvalidQuantizerGroupError(Exception):
    """
    Raised when an invalid quantizer group is encountered.
    """

    pass


class InstallationError(Exception):
    """
    Raised when an error occurs during installation.
    """

    pass


class UnsupportedModelError(Exception):
    """
    Raised when an unsupported model is encountered.
    """

    pass


class UnsupportedVersionError(Exception):
    """
    Raised when an unsupported version is encountered.
    """

    pass


class ModuleNotFoundError(Exception):
    """
    Raised when a required module is not found.
    """

    pass


class ParameterNotSupportedError(Exception):
    """
    Raised when an unsupported parameter is encountered.

    This error indicates that an unsupported parameter was provided, typically in the context
    of a compression algorithm's configuration or an argument passed to an API call.

    """

    pass


class InvalidCollectorTypeError(Exception):
    """
    Raised when an invalid collector type is encountered.
    """

    pass


class BufferFullError(Exception):
    """
    Raised when a buffer is full
    """

    pass


class UnknownDatasetError(Exception):
    """
    Raised when an unknown dataset is encountered.
    """

    pass


class StatisticsCacheError(Exception):
    """
    Raised when any error occurs during statistics caching.
    """

    def __str__(self) -> str:
        return (
            "The error occurred during statistics caching:\n"
            f"{super().__str__()}\n"
            "Please, remove the cache directory and collect cache again."
        )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from nncf.api.statistics import Statistics
from nncf.common.utils.api_marker import api
from nncf.common.utils.helpers import create_table


@api()
class SparsifiedLayerSummary:
    """
    Contains information about the sparsified layer.

    :param name: Layer's name.
    :param weight_shape: Weight's shape.
    :param sparsity_level: Sparsity level of the sparsified layer.
    :param weight_percentage: Proportion of the layer's weights in the whole model.
    """

    def __init__(
        self,
        name: str,
        weight_shape: List[int],
        sparsity_level: float,
        weight_percentage: float,
    ):
        self.name = name
        self.weight_shape = weight_shape
        self.sparsity_level = sparsity_level
        self.weight_percentage = weight_percentage


@api()
class SparsifiedModelStatistics(Statistics):
    """
    Contains statistics of the sparsified model.

    :param sparsity_level: Sparsity level of the whole model.
    :param sparsity_level_for_layers: Sparsity level of all sparsified layers
      (i.e. layers for which the algorithm was applied).
    :param sparsified_layers_summary: Detailed summary for the sparsified layers.
    """

    def __init__(
        self,
        sparsity_level: float,
        sparsity_level_for_layers: float,
        sparsified_layers_summary: List[SparsifiedLayerSummary],
    ):
        self.sparsity_level = sparsity_level
        self.sparsity_level_for_layers = sparsity_level_for_layers
        self.sparsified_layers_summary = sparsified_layers_summary

    def to_str(self) -> str:
        model_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                ["Sparsity level of the whole model", self.sparsity_level],
                [
                    "Sparsity level of all sparsified layers",
                    self.sparsity_level_for_layers,
                ],
            ],
        )

        layers_string = create_table(
            header=[
                "Layer's name",
                "Weight's shape",
                "Sparsity level",
                "Weight's percentage",
            ],
            rows=[
                [s.name, s.weight_shape, s.sparsity_level, s.weight_percentage] for s in self.sparsified_layers_summary
            ],
        )

        pretty_string = (
            f"Statistics of the sparsified model:\n{model_string}\n\n"
            f"Statistics by sparsified layers:\n{layers_string}"
        )
        return pretty_string


class LayerThreshold:
    def __init__(self, name: str, threshold: float):
        self.name = name
        self.threshold = threshold


@api()
class MagnitudeSparsityStatistics(Statistics):
    """
    Contains statistics of the magnitude sparsity algorithm.

    :param model_statistics: Statistics of the sparsified model.
    :param thresholds: List of the sparsity thresholds.
    :param target_sparsity_level: A target level of the sparsity for the algorithm for the current epoch.
    """

    def __init__(
        self,
        model_statistics: SparsifiedModelStatistics,
        thresholds: List[LayerThreshold],
        target_sparsity_level: float,
    ):
        self.model_statistics = model_statistics
        self.thresholds = thresholds
        self.target_sparsity_level = target_sparsity_level

    def to_str(self) -> str:
        thresholds_string = create_table(
            ["Layer's name", "Sparsity threshold"],
            [[s.name, s.threshold] for s in self.thresholds],
        )

        algorithm_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                [
                    "A target level of the sparsity for the algorithm for the current epoch",
                    self.target_sparsity_level,
                ],
            ],
        )

        pretty_string = (
            f"{self.model_statistics.to_str()}\n\n"
            f"Statistics of the magnitude sparsity algorithm:\n{algorithm_string}\n{thresholds_string}"
        )
        return pretty_string


@api()
class ConstSparsityStatistics(Statistics):
    """
    Contains statistics of the const sparsity algorithm.

    :param model_statistics: Statistics of the sparsified model.
    """

    def __init__(self, model_statistics: SparsifiedModelStatistics):
        self.model_statistics = model_statistics

    def to_str(self) -> str:
        pretty_string = self.model_statistics.to_str()
        return pretty_string


@api()
class RBSparsityStatistics(Statistics):
    """
    Contains statistics of the RB-sparsity algorithm.

    :param model_statistics: Statistics of the sparsified model.
    :param target_sparsity_level: A target level of the sparsity for the algorithm for the current epoch.
    :param mean_sparse_prob: The probability that one weight will be zeroed.
    """

    def __init__(
        self,
        model_statistics: SparsifiedModelStatistics,
        target_sparsity_level: float,
        mean_sparse_prob: float,
    ):
        self.model_statistics = model_statistics
        self.target_sparsity_level = target_sparsity_level
        self.mean_sparse_prob = mean_sparse_prob

    def to_str(self) -> str:
        algorithm_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                [
                    "A target level of the sparsity for the algorithm for the current epoch",
                    self.target_sparsity_level,
                ],
                [
                    "The probability that one weight will be zeroed",
                    self.mean_sparse_prob,
                ],
            ],
        )

        pretty_string = (
            f"{self.model_statistics.to_str()}\n\n Statistics of the RB-sparsity algorithm:\n{algorithm_string}"
        )
        return pretty_string


@api()
class MovementSparsityStatistics(Statistics):
    """
    Contains statistics of the movement-sparsity algorithm.

    :param model_statistics: Statistics of the sparsified model.
    :param importance_threshold: Importance threshold for sparsity binary mask.
    :param importance_regularization_factor: Penalty factor of importance score.
    """

    def __init__(
        self,
        model_statistics: SparsifiedModelStatistics,
        importance_threshold: float,
        importance_regularization_factor: float,
    ):
        self.model_statistics = model_statistics
        self.importance_threshold = importance_threshold
        self.importance_regularization_factor = importance_regularization_factor

    def to_str(self) -> str:
        algorithm_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                ["Mask Importance Threshold", self.importance_threshold],
                [
                    "Importance Regularization Factor",
                    self.importance_regularization_factor,
                ],
            ],
        )

        pretty_string = (
            f"{self.model_statistics.to_str()}\n\n"
            f"Statistics of the movement-sparsity algorithm:\n{algorithm_string}"
        )
        return pretty_string



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.patterns import HWFusedPatternNames
from nncf.common.utils.registry import Registry
from nncf.openvino.graph.metatypes import openvino_metatypes as om
from nncf.openvino.graph.metatypes.groups import ARITHMETIC_OPERATIONS
from nncf.openvino.graph.metatypes.groups import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.openvino.graph.metatypes.groups import BATCH_NORMALIZATION_OPERATIONS
from nncf.openvino.graph.metatypes.groups import ELEMENTWISE_OPERATIONS
from nncf.openvino.graph.metatypes.groups import LINEAR_OPERATIONS

OPENVINO_HW_FUSED_PATTERNS = Registry("openvino")

# BLOCK PATTERNS


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ADD_SCALE_SHIFT_OUTPUT)
def create_add_scale_shift_output() -> GraphPattern:
    pattern = GraphPattern()
    add_node_1 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    mul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    add_node_2 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    result_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_OUTPUT", GraphPattern.METATYPE_ATTR: om.OVResultMetatype}
    )

    pattern.add_edge(add_node_1, mul_node)
    pattern.add_edge(mul_node, add_node_2)
    pattern.add_edge(add_node_2, result_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.BATCH_INDEX)
def create_batch_index() -> GraphPattern:
    pattern = GraphPattern()
    subtract_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "SUBTRACT", GraphPattern.METATYPE_ATTR: om.OVSubtractMetatype}
    )
    multiply_node_1 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    multiply_node_2 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    unsqueeze_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "UNSQUEEZE", GraphPattern.METATYPE_ATTR: om.OVUnsqueezeMetatype}
    )
    # TODO (KodiaqQ): Check the pattern on real case
    concat_node_1 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    concat_node_2 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    reshape_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "RESHAPE", GraphPattern.METATYPE_ATTR: om.OVReshapeMetatype}
    )
    convolution_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONVOLUTION", GraphPattern.METATYPE_ATTR: om.OVConvolutionMetatype}
    )

    pattern.add_edge(subtract_node, multiply_node_1)
    pattern.add_edge(multiply_node_1, multiply_node_2)
    pattern.add_edge(multiply_node_2, add_node)
    pattern.add_edge(add_node, unsqueeze_node)
    pattern.add_edge(unsqueeze_node, concat_node_1)
    pattern.add_edge(concat_node_1, concat_node_2)
    pattern.add_edge(concat_node_2, reshape_node)
    pattern.add_edge(reshape_node, convolution_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.MVN_SCALE_SHIFT)
def create_mvn() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MVN", GraphPattern.METATYPE_ATTR: om.OVMVNMetatype})
    scale_shift = create_scale_shift()

    pattern.join_patterns(scale_shift)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.NORMALIZE_L2_MULTIPLY)
def create_normalize() -> GraphPattern:
    pattern = GraphPattern()
    normalize_l2_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "NORMALIZEL2", GraphPattern.METATYPE_ATTR: om.OVNormalizeL2Metatype}
    )
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(normalize_l2_node, multiply_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_WITH_BIAS)
def create_biased_op() -> GraphPattern:
    pattern = GraphPattern()
    linear_node = pattern.add_node(**{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"})
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD_BIAS", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})

    pattern.add_edge(linear_node, add_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SCALE_SHIFT)
def create_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    mul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    pattern.add_edge(mul_node, add_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SHIFT_SCALE)
def create_shift_scale() -> GraphPattern:
    pattern = GraphPattern()
    add_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, SUBTRACT",
            GraphPattern.METATYPE_ATTR: [om.OVAddMetatype, om.OVSubtractMetatype],
        }
    )
    mul_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "MULTIPLY, DIV",
            GraphPattern.METATYPE_ATTR: [om.OVMultiplyMetatype, om.OVDivideMetatype],
        }
    )
    pattern.add_edge(add_node, mul_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SOFTMAX_DIV)
def create_softmax_div() -> GraphPattern:
    pattern = GraphPattern()
    exp_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "EXP", GraphPattern.METATYPE_ATTR: om.OVExpMetatype})
    sum_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "REDUCE_SUM", GraphPattern.METATYPE_ATTR: om.OVSumMetatype})
    divide_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "DIVIDE", GraphPattern.METATYPE_ATTR: om.OVDivideMetatype}
    )

    pattern.add_edge(exp_node, sum_node)
    pattern.add_edge(exp_node, divide_node)
    pattern.add_edge(sum_node, divide_node)
    return pattern


# ACTIVATIONS


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.HSWISH_ACTIVATION)
def create_hswish() -> GraphPattern:
    pattern = GraphPattern()
    linear_node = pattern.add_node(**{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"})
    add_node_1 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD_BIAS", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    add_node_2 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    clamp_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "CLAMP", GraphPattern.METATYPE_ATTR: om.OVClampMetatype})
    multiply_node_1 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    multiply_node_2 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(linear_node, add_node_1)
    pattern.add_edge(add_node_1, add_node_2)
    pattern.add_edge(add_node_2, clamp_node)
    pattern.add_edge(add_node_1, multiply_node_1)
    pattern.add_edge(clamp_node, multiply_node_1)
    pattern.add_edge(multiply_node_1, multiply_node_2)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.HSWISH_ACTIVATION_V2)
def create_hswish_pattern_2() -> GraphPattern:
    pattern = GraphPattern()
    input_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, MULTIPLY, REDUCE_MEAN, SQUEEZE",
            GraphPattern.METATYPE_ATTR: [
                om.OVAddMetatype,
                om.OVMultiplyMetatype,
                om.OVReduceMeanMetatype,
                om.OVSqueezeMetatype,
            ],
        }
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    clamp_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "CLAMP", GraphPattern.METATYPE_ATTR: om.OVClampMetatype})
    multiply_node_1 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    multiply_node_2 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(input_node, add_node)
    pattern.add_edge(add_node, clamp_node)
    pattern.add_edge(clamp_node, multiply_node_1)
    pattern.add_edge(input_node, multiply_node_2)
    pattern.add_edge(multiply_node_1, multiply_node_2)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.HSWISH_ACTIVATION_WITHOUT_DENOMINATOR)
def create_hswish_without_denominator() -> GraphPattern:
    pattern = GraphPattern()
    linear_node = pattern.add_node(**{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"})
    add_node_1 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD_BIAS", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    add_node_2 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    clamp_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "CLAMP", GraphPattern.METATYPE_ATTR: om.OVClampMetatype})
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(linear_node, add_node_1)
    pattern.add_edge(add_node_1, add_node_2)
    pattern.add_edge(add_node_2, clamp_node)
    pattern.add_edge(add_node_1, multiply_node)
    pattern.add_edge(clamp_node, multiply_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SWISH_WITH_HARD_SIGMOID)
def create_swish_with_hardsigmoid() -> GraphPattern:
    pattern = GraphPattern()
    linear_node = pattern.add_node(**{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"})
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD_BIAS", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    hard_sigmoid_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "HARDSIGMOID", GraphPattern.METATYPE_ATTR: om.OVHardSigmoidMetatype}
    )
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(linear_node, add_node)
    pattern.add_edge(add_node, hard_sigmoid_node)
    pattern.add_edge(add_node, multiply_node)
    pattern.add_edge(hard_sigmoid_node, multiply_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SOFTMAX)
def create_softmax() -> GraphPattern:
    pattern = GraphPattern()
    exp_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "EXP", GraphPattern.METATYPE_ATTR: om.OVExpMetatype})
    sum_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "REDUCE_SUM", GraphPattern.METATYPE_ATTR: om.OVSumMetatype})
    power_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "POWER", GraphPattern.METATYPE_ATTR: om.OVPowerMetatype})
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(exp_node, sum_node)
    pattern.add_edge(sum_node, power_node)
    pattern.add_edge(exp_node, multiply_node)
    pattern.add_edge(power_node, multiply_node)
    return pattern


# INPUT PROCESSING


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_CONVERT_TRANSPOSE_PROCESSING)
def create_input_convert_transpose_processing() -> GraphPattern:
    input_convert_transpose = create_input_convert_transpose()
    pattern = GraphPattern()
    pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, MULTIPLY, SUBTRACT",
            GraphPattern.METATYPE_ATTR: [om.OVAddMetatype, om.OVMultiplyMetatype, om.OVSubtractMetatype],
        }
    )

    input_convert_transpose.join_patterns(pattern)
    return input_convert_transpose


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_CONVERT_TRANSPOSE_REVERSE_ADD)
def create_input_convert_transpose_reverse_add() -> GraphPattern:
    input_convert_transpose = create_input_convert_transpose()
    pattern = GraphPattern()
    split_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SPLIT", GraphPattern.METATYPE_ATTR: om.OVSplitMetatype})
    # TODO (KodiaqQ): Check the pattern on real case
    concat_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})

    pattern.add_edge(split_node, concat_node)
    pattern.add_edge(concat_node, add_node)
    input_convert_transpose.join_patterns(pattern)
    return input_convert_transpose


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_CONVERT_TRANSPOSE_REVERSE_SCALE_SHIFT)
def create_input_convert_transpose_reverse_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    convert_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONVERT", GraphPattern.METATYPE_ATTR: om.OVConvertMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.OVTransposeMetatype}
    )
    split_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SPLIT", GraphPattern.METATYPE_ATTR: om.OVSplitMetatype})
    # TODO (KodiaqQ): Check the pattern on real case
    concat_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    scale_shift = create_scale_shift()

    pattern.add_edge(model_input, convert_node)
    pattern.add_edge(convert_node, transpose_node)
    pattern.add_edge(transpose_node, split_node)
    pattern.add_edge(split_node, concat_node)
    pattern.join_patterns(scale_shift)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_CONVERT_TRANSPOSE_SCALE_SHIFT)
def create_input_convert_transpose_scale_shift() -> GraphPattern:
    input_convert_transpose = create_input_convert_transpose()
    scale_shift = create_scale_shift()
    input_convert_transpose.join_patterns(scale_shift)
    return input_convert_transpose


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_PROCESSING)
def create_input_processing() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    processing_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "SUBTRACT, MULTIPLY, ADD",
            GraphPattern.METATYPE_ATTR: [om.OVSubtractMetatype, om.OVMultiplyMetatype, om.OVAddMetatype],
        }
    )

    pattern.add_edge(model_input, processing_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_REVERSE_ADD)
def create_input_reverse_add() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    split_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SPLIT", GraphPattern.METATYPE_ATTR: om.OVSplitMetatype})
    # TODO (KodiaqQ): Check the pattern on real case
    concat_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})

    pattern.add_edge(model_input, split_node)
    pattern.add_edge(split_node, concat_node)
    pattern.add_edge(concat_node, add_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_REVERSE_SCALE_SHIFT)
def create_input_reverse_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    split_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SPLIT", GraphPattern.METATYPE_ATTR: om.OVSplitMetatype})
    # TODO (KodiaqQ): Check the pattern on real case
    concat_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    scale_shift = create_scale_shift()

    pattern.add_edge(model_input, split_node)
    pattern.add_edge(split_node, concat_node)
    pattern.join_patterns(scale_shift)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_SCALE_SHIFT)
def create_input_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype})
    scale_shift = create_scale_shift()

    pattern.join_patterns(scale_shift)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_SHIFT_SCALE)
def create_input_shift_scale() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype})
    shift_scale = create_shift_scale()

    pattern.join_patterns(shift_scale)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_TRANSPOSE_PROCESSING)
def create_input_transpose_processing() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.OVTransposeMetatype}
    )
    processing_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, MULTIPLY, SUBTRACT",
            GraphPattern.METATYPE_ATTR: [om.OVAddMetatype, om.OVMultiplyMetatype, om.OVSubtractMetatype],
        }
    )

    pattern.add_edge(model_input, transpose_node)
    pattern.add_edge(transpose_node, processing_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_TRANSPOSE_REVERSE_ADD)
def create_input_transpose_reverse_add() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.OVTransposeMetatype}
    )
    split_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SPLIT", GraphPattern.METATYPE_ATTR: om.OVSplitMetatype})
    # TODO (KodiaqQ): Check the pattern on real case
    concat_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.OVConcatMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})

    pattern.add_edge(model_input, transpose_node)
    pattern.add_edge(transpose_node, split_node)
    pattern.add_edge(split_node, concat_node)
    pattern.add_edge(concat_node, add_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_TRANSPOSE_SCALE_SHIFT)
def create_input_transpose_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.OVTransposeMetatype}
    )
    scale_shift = create_scale_shift()

    pattern.add_edge(model_input, transpose_node)
    pattern.join_patterns(scale_shift)
    return pattern


# COMBINATIONS


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ACTIVATIONS_BATCH_NORM)
def create_activations_batch_norm() -> GraphPattern:
    activations = atomic_activations_operations()
    batch_norm = batch_normalization_operations()
    activations.join_patterns(batch_norm)
    return activations


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS)
def create_arithmetic_activations() -> GraphPattern:
    arithmetic = arithmetic_operations()
    activations = atomic_activations_operations()
    arithmetic.join_patterns(activations)
    return arithmetic


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS_ARITHMETIC)
def create_arithmetic_activations_arithmetic() -> GraphPattern:
    arithmetic_activations = create_arithmetic_activations()
    arithmetic = arithmetic_operations()
    arithmetic_activations.join_patterns(arithmetic)
    return arithmetic_activations


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.BATCH_NORM_ACTIVATIONS)
def create_batch_norm_activations() -> GraphPattern:
    batch_norm = batch_normalization_operations()
    activations = atomic_activations_operations()
    batch_norm.join_patterns(activations)
    return batch_norm


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS)
def create_linear_activations() -> GraphPattern:
    linear = linear_operations()
    activations = atomic_activations_operations()
    linear.join_patterns(activations)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC)
def create_linear_arithmetic() -> GraphPattern:
    linear = linear_operations()
    arithmetic = arithmetic_operations()
    linear.join_patterns(arithmetic)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC_ACTIVATIONS)
def create_linear_arithmetic_activations() -> GraphPattern:
    linear = linear_operations()
    arithmetic = arithmetic_operations()
    activations = atomic_activations_operations()

    linear.join_patterns(arithmetic)
    linear.join_patterns(activations)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SHIFT_SCALE)
def create_linear_shift_scale() -> GraphPattern:
    linear = linear_operations()
    shift_scale = create_shift_scale()
    linear.join_patterns(shift_scale)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC_ACTIVATIONS_ARITHMETIC)
def create_linear_arithmetic_activations_arithmetic() -> GraphPattern:
    linear_arithmetic_activations = create_linear_arithmetic_activations()
    arithmetic = arithmetic_operations()

    linear_arithmetic_activations.join_patterns(arithmetic)
    return linear_arithmetic_activations


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SQUEEZE_ACTIVATIONS)
def create_linear_squeeze_activation() -> GraphPattern:
    linear = linear_operations()
    squeeze = squeeze_operation()
    activations = atomic_activations_operations()

    linear.join_patterns(squeeze)
    linear.join_patterns(activations)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SQUEEZE_ARITHMETIC_ACTIVATIONS)
def create_linear_squeeze_arithmetic_activation() -> GraphPattern:
    linear = linear_operations()
    squeeze = squeeze_operation()
    arithmetic_activations = create_arithmetic_activations()

    linear.join_patterns(squeeze)
    linear.join_patterns(arithmetic_activations)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.MVN_SCALE_SHIFT_ACTIVATIONS)
def create_mvn_scale_shift_activations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MVN", GraphPattern.METATYPE_ATTR: om.OVMVNMetatype})
    scale_shift = create_scale_shift()
    activations = atomic_activations_operations()

    pattern.join_patterns(scale_shift)
    pattern.join_patterns(activations)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS_UNSQUEEZE_BN_SQUEEZE)
def create_linear_activations_unsqueeze_bn_squeeze():
    linear_biased = create_biased_op()
    activations = atomic_activations_operations()
    unsqueeze_op = unsqueeze_operation()
    scale_shift = create_scale_shift()
    squeeze_op = squeeze_operation()

    linear_biased.join_patterns(activations)
    linear_biased.join_patterns(unsqueeze_op)
    linear_biased.join_patterns(scale_shift)
    linear_biased.join_patterns(squeeze_op)
    return linear_biased


# DEVICE PATTERNS


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.HSWISH_ACTIVATION_CLAMP_MULTIPLY)
def create_clamp_mult_const() -> GraphPattern:
    pattern = GraphPattern()
    clamp_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "CLAMP", GraphPattern.METATYPE_ATTR: om.OVClampMetatype})
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )

    pattern.add_edge(clamp_node, multiply_node)
    return pattern


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SCALE_SHIFT_ACTIVATIONS)
def create_scale_shift_activations() -> GraphPattern:
    scale_shift = create_scale_shift()
    activations = atomic_activations_operations()
    scale_shift.join_patterns(activations)
    return scale_shift


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SCALE_SHIFT)
def create_linear_scale_shift() -> GraphPattern:
    linear = linear_operations()
    scale_shift = create_scale_shift()
    linear.join_patterns(scale_shift)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BIASED_SCALE_SHIFT)
def create_linear_biased_scale_shift() -> GraphPattern:
    linear_biased = create_biased_op()
    scale_shift = create_scale_shift()
    linear_biased.join_patterns(scale_shift)
    return linear_biased


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATION_SCALE_SHIFT)
def create_linear_activation_scale_shift() -> GraphPattern:
    linear_activations = create_linear_activations()
    scale_shift = create_scale_shift()

    linear_activations.join_patterns(scale_shift)
    return linear_activations


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BIASED_ACTIVATION_SCALE_SHIFT)
def create_linear_biased_activation_scale_shift() -> GraphPattern:
    linear_biased = create_biased_op()
    activations = atomic_activations_operations()
    scale_shift = create_scale_shift()

    linear_biased.join_patterns(activations)
    linear_biased.join_patterns(scale_shift)
    return linear_biased


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_TO_SPACE_SCALE_SHIFT_ACTIVATIONS)
def create_linear_batch_to_space_scale_shift_activations() -> GraphPattern:
    linear = linear_operations()
    batch_to_space = batch_to_space_operation()
    scale_shift = create_scale_shift()
    activations = atomic_activations_operations()

    linear.join_patterns(batch_to_space)
    linear.join_patterns(scale_shift)
    linear.join_patterns(activations)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_TO_SPACE_ARITHMETIC_ACTIVATIONS)
def create_linear_batch_to_space_arithmetic_activations() -> GraphPattern:
    linear = linear_operations()
    batch_to_space = batch_to_space_operation()
    arithmetic = arithmetic_operations()
    activations = atomic_activations_operations()

    linear.join_patterns(batch_to_space)
    linear.join_patterns(arithmetic)
    linear.join_patterns(activations)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ELEMENTWISE)
def create_linear_elementwise() -> GraphPattern:
    linear = linear_operations()
    elementwise = elementwise_operations()
    linear.join_patterns(elementwise)
    return linear


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BIASED_ELEMENTWISE)
def create_linear_biased_elementwise() -> GraphPattern:
    linear_biased = create_biased_op()
    elementwise = elementwise_operations()
    linear_biased.join_patterns(elementwise)
    return linear_biased


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATION_ELEMENTWISE)
def create_linear_activation_elementwise() -> GraphPattern:
    linear_activations = create_linear_activations()
    elementwise = elementwise_operations()

    linear_activations.join_patterns(elementwise)
    return linear_activations


@OPENVINO_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BIASED_ACTIVATION_ELEMENTWISE)
def create_linear_biased_activation_elementwise() -> GraphPattern:
    linear_biased = create_biased_op()
    activations = atomic_activations_operations()
    elementwise = elementwise_operations()

    linear_biased.join_patterns(activations)
    linear_biased.join_patterns(elementwise)
    return linear_biased


def elementwise_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.METATYPE_ATTR: ELEMENTWISE_OPERATIONS, GraphPattern.LABEL_ATTR: "ELEMENTWISE"})
    return pattern


def linear_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"})
    return pattern


def batch_normalization_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(
        **{GraphPattern.METATYPE_ATTR: BATCH_NORMALIZATION_OPERATIONS, GraphPattern.LABEL_ATTR: "BATCH_NORMALIZATION"}
    )
    return pattern


def atomic_activations_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(
        **{GraphPattern.METATYPE_ATTR: ATOMIC_ACTIVATIONS_OPERATIONS, GraphPattern.LABEL_ATTR: "ATOMIC_ACTIVATIONS"}
    )
    return pattern


def arithmetic_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.METATYPE_ATTR: ARITHMETIC_OPERATIONS, GraphPattern.LABEL_ATTR: "ARITHMETIC"})
    return pattern


def squeeze_operation() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "SQUEEZE", GraphPattern.METATYPE_ATTR: om.OVSqueezeMetatype})
    return pattern


def unsqueeze_operation() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "UNSQUEEZE", GraphPattern.METATYPE_ATTR: om.OVUnsqueezeMetatype})
    return pattern


def batch_to_space_operation() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "BATCH_TO_SPACE", GraphPattern.METATYPE_ATTR: om.OVBatchToSpaceMetatype}
    )
    return pattern


def create_input_convert_transpose() -> GraphPattern:
    pattern = GraphPattern()
    model_input = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: om.OVParameterMetatype}
    )
    convert_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONVERT", GraphPattern.METATYPE_ATTR: om.OVConvertMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.OVTransposeMetatype}
    )
    pattern.add_edge(model_input, convert_node)
    pattern.add_edge(convert_node, transpose_node)
    return pattern



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

import openvino.runtime as ov

from nncf.common.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.data.dataset import Dataset
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.statistics.collectors import get_raw_stat_collector
from nncf.quantization.algorithms.layerwise.backend import LayerwiseEngineBackend
from nncf.quantization.algorithms.layerwise.openvino_iterator import OVLayerwiseIterator
from nncf.quantization.algorithms.layerwise.scheduler import LayerwiseStep
from nncf.quantization.algorithms.layerwise.scheduler import NodeOutputPort
from nncf.tensor import Tensor


class OVLayerwiseEngineBackend(LayerwiseEngineBackend):
    @staticmethod
    def create_layerwise_iterator(
        model: ov.Model,
        graph: NNCFGraph,
        schedule: List[LayerwiseStep],
        dataset: Dataset,
        subset_size: int = 100,
        cache: Optional[Dict[NodeOutputPort, List[Tensor]]] = None,
    ) -> OVLayerwiseIterator:
        return OVLayerwiseIterator(model, graph, schedule, dataset, subset_size, cache)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorStatisticCollectorBase:
        return get_raw_stat_collector(num_samples)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from typing import Dict, Optional, Set

from nncf.common.quantization.quantizer_propagation.structs import PropagatingQuantizer


class UnifiedScalePropagatingQuantizerGroupManager:
    """
    Keeps track of the groups of quantizers that have to have their scales unified in the final
    quantized model.
    """

    def __init__(self) -> None:
        self._next_gid = 0
        self._group_vs_prop_quants_dict: Dict[int, Set[PropagatingQuantizer]] = {}

    def _get_next_gid(self) -> int:
        retval = self._next_gid
        self._next_gid += 1
        return retval

    def register_group(self, prop_quants: Set[PropagatingQuantizer]) -> int:
        """
        Registers a set of propagating quantizers as a new group.

        :param prop_quants: A set of propagating quantizers to be registered.
        :return: The ID of the newly created group.
        """
        for pq in prop_quants:
            for gid, group in self._group_vs_prop_quants_dict.items():
                assert pq not in group, "Propagating quantizer #{} is already registered in a group {}!".format(
                    pq.id, gid
                )
        gid = self._get_next_gid()
        self._group_vs_prop_quants_dict[gid] = prop_quants
        return gid

    def add_to_group(self, target_gid: int, prop_quant: PropagatingQuantizer) -> None:
        """
        Adds a propagating quantizer to an already existing group.

        :param target_gid: The ID of the group to be extended.
        :param prop_quant: The propagating quantizer to be registered in the group. The quantizer
          must not be already registered in any group.
        """
        for gid, group in self._group_vs_prop_quants_dict.items():
            if target_gid != gid:
                assert prop_quant not in group, (
                    "Tried to add propagating quantizer #{} to group #{}, "
                    "but it is already registered in a group {}!".format(prop_quant.id, target_gid, gid)
                )
        self._group_vs_prop_quants_dict[target_gid].add(prop_quant)

    def remove_from_group(self, group: int, prop_quant: PropagatingQuantizer) -> None:
        """
        Removes a propagating quantizer from a group.

        :param group: The ID of the group from where a quantizer should be removed.
        :param prop_quant: The propagating quantizer to be removed from the group.
        """
        self._group_vs_prop_quants_dict[group].remove(prop_quant)

    def get_group_vs_prop_quants_dict(self) -> Dict[int, Set[PropagatingQuantizer]]:
        """
        :return: A dictionary of groups vs propagating quantizers currently associated with the corresponding group.
        """
        return copy(self._group_vs_prop_quants_dict)

    def get_group_id_by_propagating_quantizer_id(self, requested_pqid: int) -> Optional[int]:
        """
        If a propagating quantizer with a given ID is registered within a group,
        then this function will return the corresponding group ID; otherwise, None is returned.

        :param requested_pqid: The ID of the propagating quantizer to search for among groups.
        :return: The group ID of the quantizer, if found, otherwise None.
        """
        for gid, group in self._group_vs_prop_quants_dict.items():
            for pq in group:
                if pq.id == requested_pqid:
                    return gid
        return None

    def merge_groups(self, merge_to_gid: int, merge_from_gid: int) -> None:
        """
        Merges two groups into a single one. The `merge_to_gid` group retains its group ID.

        :param merge_to_gid: The ID of the group to merge into.
        :param merge_from_gid: The ID of the group to merge into the group defined by `merge_to_gid`
        """
        if merge_to_gid == merge_from_gid:
            return
        self._group_vs_prop_quants_dict[merge_to_gid].update(self._group_vs_prop_quants_dict[merge_from_gid])
        self._group_vs_prop_quants_dict.pop(merge_from_gid)


class QuantizersWaitingForMergeManager:
    """
    Tracks the quantizers that await a merge while trying to transition through a downward-branching node
    and corresponding node keys.
    """

    def __init__(self) -> None:
        self._branching_node_keys_vs_quantizers_waiting_for_merge: Dict[str, Set[PropagatingQuantizer]] = {}
        self._quantizers_vs_branching_node_keys: Dict[PropagatingQuantizer, str] = {}

    def add_propagating_quantizer_to_wait_on_node_key(self, pq: PropagatingQuantizer, branching_node_key: str) -> None:
        """
        Registers a propagating quantizer as "waiting" on a node in QuantizerPropagationStateGraph.

        :param pq: The propagating quantizer to be registered as "waiting".
        :param branching_node_key: The node key in QuantizerPropagationStateGraph to be waited upon, most likely
          the downward-branching node.
        """
        if branching_node_key not in self._branching_node_keys_vs_quantizers_waiting_for_merge:
            self._branching_node_keys_vs_quantizers_waiting_for_merge[branching_node_key] = set()
        self._branching_node_keys_vs_quantizers_waiting_for_merge[branching_node_key].add(pq)
        self._quantizers_vs_branching_node_keys[pq] = branching_node_key

    def get_blocking_node(self, pq: PropagatingQuantizer) -> str:
        """
        Returns the node key upon which the propagating quantizer is registered to be "waiting".

        :param pq: The propagating quantizer that has already been registered to be "waiting" on a node.
        :return: The node key in QuantizerPropagationStateGraph that the `pq` is registered to be waiting upon.
        """
        return self._quantizers_vs_branching_node_keys[pq]

    def get_waiting_quantizers_for_branching_node_key(self, node_key: str) -> Set[PropagatingQuantizer]:
        """
        Returns the set of all quantizers registered to be "waiting" on a given node key in
        QuantizerPropagationStateGraph.

        :param node_key: The node key in QuantizerPropagationStateGraph
        :return: The set of propagating quantizers registered as "waiting" on `node_key`.
        """
        return self._branching_node_keys_vs_quantizers_waiting_for_merge[node_key]

    def __contains__(self, item: PropagatingQuantizer) -> bool:
        return item in self._quantizers_vs_branching_node_keys

    def resolve_merged_node(self, branching_node_key: str) -> None:
        """
        De-registers any quantizers that were previously registered to be "waiting" on a given node key.
        :param branching_node_key: The node key in QuantizerPropagationStateGraph that some propagating
          quantizers have previously been registered upon.
        """
        for pq in self._branching_node_keys_vs_quantizers_waiting_for_merge[branching_node_key]:
            self._quantizers_vs_branching_node_keys.pop(pq)
        self._branching_node_keys_vs_quantizers_waiting_for_merge.pop(branching_node_key)
